import numpy as np
from teleop import Teleop


def main():
    def callback(pose, message):
        print(f"Pose: {pose}")
        print(f"Message: {message}")

    teleop = Teleop()
    teleop.set_pose(np.eye(4))
    teleop.subscribe(callback)
    teleop.run()


if __name__ == "__main__":
    main()



import ssl
import os
import math
import logging
from werkzeug.serving import ThreadedWSGIServer
from typing import Callable
from flask import Flask, send_from_directory, request
import transforms3d as t3d
import numpy as np


TF_RUB2FLU = np.array([[0, 0, -1, 0], [-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
THIS_DIR = os.path.dirname(os.path.realpath(__file__))


def are_close(a, b=None, lin_tol=1e-9, ang_tol=1e-9):
    """
    Check if two transformation matrices are close to each other within specified tolerances.

    Parameters:
        a (numpy.ndarray): The first transformation matrix.
        b (numpy.ndarray, optional): The second transformation matrix. If not provided, it defaults to the identity matrix.
        lin_tol (float, optional): The linear tolerance for closeness. Defaults to 1e-9.
        ang_tol (float, optional): The angular tolerance for closeness. Defaults to 1e-9.

    Returns:
        bool: True if the matrices are close, False otherwise.
    """
    if b is None:
        b = np.eye(4)
    d = np.linalg.inv(a) @ b
    if not np.allclose(d[:3, 3], np.zeros(3), atol=lin_tol):
        return False
    rpy = t3d.euler.mat2euler(d[:3, :3])
    return np.allclose(rpy, np.zeros(3), atol=ang_tol)


class Teleop:
    """
    Teleop class for controlling a robot remotely.

    Args:
        host (str, optional): The host IP address. Defaults to "0.0.0.0".
        port (int, optional): The port number. Defaults to 4443.
        ssl_context (ssl.SSLContext, optional): The SSL context for secure communication. Defaults to None.
    """

    def __init__(self, host="0.0.0.0", port=4443, ssl_context=None):
        self.__logger = logging.getLogger("teleop")
        self.__logger.setLevel(logging.INFO)
        self.__logger.addHandler(logging.StreamHandler())

        self.__server = None
        self.__host = host
        self.__port = port
        self.__ssl_context = ssl_context

        self.__relative_pose_init = None
        self.__absolute_pose_init = None
        self.__previous_received_pose = None
        self.__callbacks = []
        self.__pose = np.eye(4)

        if self.__ssl_context is None:
            self.__ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
            self.__ssl_context.load_cert_chain(
                certfile=os.path.join(THIS_DIR, "cert.pem"),
                keyfile=os.path.join(THIS_DIR, "key.pem"),
            )

        self.__app = Flask(__name__)
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)
        self.__register_routes()

    def set_pose(self, pose: np.ndarray) -> None:
        """
        Set the current pose of the end-effector.

        Parameters:
        - pose (np.ndarray): A 4x4 transformation matrix representing the pose.
        """
        self.__pose = pose

    def subscribe(self, callback: Callable[[np.ndarray, dict], None]) -> None:
        """
        Subscribe to receive updates from the teleop module.

        Parameters:
            callback (Callable[[np.ndarray, dict], None]): A callback function that will be called when pose updates are received.
                The callback function should take two arguments:
                    - np.ndarray: A 4x4 transformation matrix representing the end-effector target pose.
                    - dict: A dictionary containing additional information.
        """
        self.__callbacks.append(callback)

    def __notify_subscribers(self, pose, message):
        for callback in self.__callbacks:
            callback(pose, message)

    def __update(self, message):
        move = message["move"]
        position = message["position"]
        orientation = message["orientation"]
        reference_frame = message["reference_frame"]

        if reference_frame not in ["gripper", "base"]:
            raise ValueError(
                f'Unknown reference frame: {reference_frame} (should be "gripper" or "base")'
            )

        position = np.array([position["x"], position["y"], position["z"]])
        quat = np.array(
            [orientation["w"], orientation["x"], orientation["y"], orientation["z"]]
        )

        if not move:
            self.__relative_pose_init = None
            self.__absolute_pose_init = None
            self.__notify_subscribers(self.__pose, message)
            return

        received_pose_rub = t3d.affines.compose(
            position, t3d.quaternions.quat2mat(quat), [1, 1, 1]
        )
        received_pose = TF_RUB2FLU @ received_pose_rub
        received_pose[:3, :3] = received_pose[:3, :3] @ np.linalg.inv(
            TF_RUB2FLU[:3, :3]
        )

        # Pose jump protection
        if self.__previous_received_pose is not None:
            if not are_close(
                received_pose,
                self.__previous_received_pose,
                lin_tol=6e-2,
                ang_tol=math.radians(25),
            ):
                self.__logger.warning("Pose jump detected, resetting the pose")
                self.__relative_pose_init = None
                self.__previous_received_pose = received_pose
                return
        self.__previous_received_pose = received_pose

        # Accumulate the pose and publish
        if self.__relative_pose_init is None:
            self.__relative_pose_init = received_pose
            self.__absolute_pose_init = self.__pose
            self.__previous_received_pose = None

        relative_pose = np.linalg.inv(self.__relative_pose_init) @ received_pose
        if reference_frame == "gripper":
            self.__pose = self.__absolute_pose_init @ relative_pose
        else:
            self.__pose = np.eye(4)
            self.__pose[:3, 3] = self.__absolute_pose_init[:3, 3] + relative_pose[:3, 3]
            self.__pose[:3, :3] = (
                relative_pose[:3, :3] @ self.__absolute_pose_init[:3, :3]
            )

        # Notify the subscribers
        self.__notify_subscribers(self.__pose, message)

    def __register_routes(self):
        @self.__app.route("/<path:filename>")
        def serve_file(filename):
            self.__logger.debug(f"Serving the {filename} file")
            return send_from_directory(THIS_DIR, filename)

        @self.__app.route("/pose", methods=["POST"])
        def pose():
            json_data = request.get_json()
            self.__logger.debug(f"Received a pose update request: {json_data}")
            self.__update(json_data)
            return {"status": "ok"}

        @self.__app.route("/log", methods=["POST"])
        def log():
            json_data = request.get_json()
            self.__logger.info(f"Received a log message: {json_data}")
            return {"status": "ok"}

        @self.__app.route("/")
        def index():
            self.__logger.debug("Serving the index.html file")
            return send_from_directory(THIS_DIR, "index.html")

    def run(self) -> None:
        """
        Runs the teleop server. This method is blocking.
        """
        self.__logger.info(f"Server started at https://{self.__host}:{self.__port}")
        self.__server = ThreadedWSGIServer(
            app=self.__app,
            host=self.__host,
            port=self.__port,
            ssl_context=self.__ssl_context,
        )
        self.__server.serve_forever()

    def stop(self) -> None:
        """
        Stops the teleop server.
        """
        self.__server.shutdown()



import unittest
import threading
import requests
import time
from teleop import Teleop


def get_message():
    return {
        "move": False,
        "position": {"x": 0.0, "y": 0.0, "z": 0.0},
        "orientation": {"w": 1.0, "x": 0.0, "y": 0.0, "z": 0.0},
        "reference_frame": "base",
    }


BASE_URL = "https://localhost:4443"


class TestPoseCompounding(unittest.TestCase):
    @classmethod
    def __callback(cls, pose, message):
        cls.__last_pose = pose
        cls.__last_message = message
        print(pose)

    @classmethod
    def setUpClass(cls):
        cls.__last_pose = None
        cls.__last_message = None

        cls.teleop = Teleop()
        cls.teleop.subscribe(cls.__callback)
        cls.thread = threading.Thread(target=cls.teleop.run)
        cls.thread.daemon = True
        cls.thread.start()

    def test_response(self):
        payload = get_message()
        response = requests.post(
            f"{BASE_URL}/pose", json=payload, verify=False, timeout=5
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"status": "ok"})

    def test_single_position_update(self):
        payload = get_message()

        # The first message with `move==True` is used as a reference
        payload["move"] = True
        response = requests.post(
            f"{BASE_URL}/pose", json=payload, verify=False, timeout=5
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(self.__last_pose)
        self.assertIsNotNone(self.__last_message)

        # Move the phone up by 5cm (Y-axis)
        payload["move"] = True
        payload["position"]["y"] = 0.05
        response = requests.post(
            f"{BASE_URL}/pose", json=payload, verify=False, timeout=5
        )
        self.assertEqual(response.status_code, 200)

        # In total the result should be 5cm on the Z-axis because of the RUB -> FLU conversion
        self.assertEqual(self.__last_pose[2, 3], 0.05)

        # Move the phone up by another 5cm (Y-axis)
        payload["move"] = True
        payload["position"]["y"] = 0.1
        response = requests.post(
            f"{BASE_URL}/pose", json=payload, verify=False, timeout=5
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(self.__last_pose[2, 3], 0.1)

    @classmethod
    def tearDownClass(cls):
        cls.teleop.stop()


if __name__ == "__main__":
    unittest.main()



from teleop import Teleop
import transforms3d as t3d
import argparse

try:
    import rclpy
    import geometry_msgs.msg
    from geometry_msgs.msg import PoseStamped, TransformStamped
    from tf2_ros import TransformBroadcaster
except ImportError:
    raise ImportError(
        "ROS2 is not sourced. Please source ROS2 before running this script."
    )


def ros2numpy(pose):
    xyz = None
    q = None

    if isinstance(pose, geometry_msgs.msg.PoseStamped):
        pose = pose.pose
    elif isinstance(pose, geometry_msgs.msg.TransformStamped):
        pose = pose.transform

    if isinstance(pose, geometry_msgs.msg.Pose):
        xyz = [pose.position.x, pose.position.y, pose.position.z]
        q = [
            pose.orientation.w,
            pose.orientation.x,
            pose.orientation.y,
            pose.orientation.z,
        ]
    elif isinstance(pose, geometry_msgs.msg.Transform):
        xyz = [pose.translation.x, pose.translation.y, pose.translation.z]
        q = [pose.rotation.w, pose.rotation.x, pose.rotation.y, pose.rotation.z]
    else:
        raise ValueError("Unknown type")

    transform = t3d.affines.compose(xyz, t3d.quaternions.quat2mat(q), [1, 1, 1])
    return transform


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--omit-current-pose", action="store_true", help="Omit usage of current pose"
    )
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Host address")
    parser.add_argument("--port", type=int, default=4443, help="Port number")
    parser.add_argument(
        "--ros-args",
        nargs=argparse.REMAINDER,
        help="Arguments to pass to ROS",
    )

    args = parser.parse_args()

    rclpy.init(args=["--ros-args"] + args.ros_args)

    teleop = Teleop(host=args.host, port=args.port)
    current_robot_pose_message = None
    if args.omit_current_pose:
        current_robot_pose_message = PoseStamped()
        current_robot_pose_message.pose.orientation.w = 1
    pose_initiated = False
    node = rclpy.create_node("ros2_teleop")
    pose_publisher = node.create_publisher(PoseStamped, "target_frame", 1)
    broadcaster = TransformBroadcaster(node)

    def ros_current_pose_callback(msg):
        nonlocal current_robot_pose_message
        current_robot_pose_message = msg

    def teleop_pose_callback(pose, params):
        nonlocal current_robot_pose_message
        nonlocal teleop
        nonlocal node
        nonlocal pose_publisher
        nonlocal pose_initiated

        tf_message = TransformStamped()
        tf_message.header.stamp = node.get_clock().now().to_msg()
        tf_message.header.frame_id = "base_link"
        tf_message.child_frame_id = "teleop_target"
        tf_message.transform.translation.x = pose[0, 3]
        tf_message.transform.translation.y = pose[1, 3]
        tf_message.transform.translation.z = pose[2, 3]
        quat = t3d.quaternions.mat2quat(pose[:3, :3])
        tf_message.transform.rotation.w = quat[0]
        tf_message.transform.rotation.x = quat[1]
        tf_message.transform.rotation.y = quat[2]
        tf_message.transform.rotation.z = quat[3]
        broadcaster.sendTransform(tf_message)

        try:
            rclpy.spin_once(node, timeout_sec=0.01)
        except Exception as e:
            pass

        if current_robot_pose_message is None:
            return

        if not params["move"] and not pose_initiated:
            current_robot_pose = ros2numpy(current_robot_pose_message.pose)
            teleop.set_pose(current_robot_pose)
            pose_initiated = True
            return

        message = PoseStamped()
        message.header.stamp = node.get_clock().now().to_msg()
        message.header.frame_id = "link_base"
        message.pose.position.x = pose[0, 3]
        message.pose.position.y = pose[1, 3]
        message.pose.position.z = pose[2, 3]
        quat = t3d.quaternions.mat2quat(pose[:3, :3])
        message.pose.orientation.w = quat[0]
        message.pose.orientation.x = quat[1]
        message.pose.orientation.y = quat[2]
        message.pose.orientation.z = quat[3]
        pose_publisher.publish(message)

    current_pose_subscriber = node.create_subscription(
        PoseStamped, "/current_pose", ros_current_pose_callback, 1
    )

    teleop.subscribe(teleop_pose_callback)
    teleop.run()


if __name__ == "__main__":
    main()



import re
from pathlib import Path
from setuptools import setup


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# Replace relative links with absolute links
for match in re.findall(r"\]\((?!http)([^)]+)\)", long_description):
    filepath = Path(match)
    long_description = long_description.replace(
        match, f"https://github.com/SpesRobotics/teleop/raw/main/{filepath}"
    )


setup(
    name="teleop",
    version="0.0.6",
    packages=["teleop", "teleop.basic", "teleop.ros2"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="Turns your phone into a robot arm teleoperation device by leveraging the WebXR API",
    install_requires=[
        "Flask",
        "numpy",
        "transforms3d",
        "werkzeug",
        "pytest",
        "requests",
    ],
    package_data={
        "teleop": ["cert.pem", "key.pem", "index.html"],
    },
    license="Apache 2.0",
    author="Spes Robotics",
    author_email="contact@spes.ai",
    project_urls={
        "Documentation": "https://github.com/SpesRobotics/teleop",
        "Source": "https://github.com/SpesRobotics/teleop",
        "Tracker": "https://github.com/SpesRobotics/teleop/issues",
    },
)
